from collections.abc import Callable,Iterator
from ledsign.backend import LEDSignProtocolError
from ledsign.checksum import LEDSignCRC
from ledsign.keypoint_list import LEDSignKeypoint,LEDSignKeypointList
from ledsign.program_io import LEDSignCompiledProgram,LEDSignProgramParser
from ledsign.protocol import LEDSignProtocol
from typing import Union
import ledsign.device
import os
import struct
import sys
import threading
import weakref



__all__=["LEDSignProgramError","LEDSignProgram","LEDSignProgramBuilder"]



class LEDSignProgramError(Exception):
	"""
	Raised whenever issues in a program's configuration are detected.
	"""



class LEDSignProgram(object):
	"""
	Contains information about a complete LED sign program compatible with :python:`device`. If the :python:`file_path` argument is given, the program is loaded from the specified file path.

	An instance of this class can be used as a function decorator to generate programs dynamically (see :py:class:`LEDSignProgramBuilder` or :py:func:`__call__` for details).
	"""

	error_output_file=sys.stderr

	__slots__=["_hardware","_duration","_keypoint_list","_load_parameters","_builder_ready","_has_error"]

	def __init__(self,device:"LEDSign",file_path:str|None=None) -> None:
		if (not isinstance(device,ledsign.device.LEDSign)):
			raise TypeError(f"Expected 'LEDSign', got '{device.__class__.__name__}'")
		if (file_path is not None and not isinstance(file_path,str)):
			raise TypeError(f"Expected 'str', got '{file_path.__class__.__name__}'")
		self._hardware=device._hardware
		self._duration=1
		self._keypoint_list=LEDSignKeypointList()
		self._load_parameters=None
		self._builder_ready=False
		self._has_error=False
		if (file_path is not None):
			self._load_from_file(file_path)

	def __repr__(self) -> str:
		return f"<LEDSignProgram{('[unloaded]' if self._load_parameters is not None else '')} hardware={self._hardware.get_string()} duration={self._duration/60:.3f}s>"

	def __call__(self,func:Callable[[],None],args:tuple|list=(),kwargs:dict={},inject_commands:bool=True,bypass_errors:bool=False) -> "LEDSignProgram":
		"""
		Explicitly generates a program from the given function :python:`func(*args, **kwargs)`, and optionally bypasses error verification if the :python:`bypass_errors` flag is set. Explicit uses of this method are required when 'stitching' a program from multiple functions, or when it is necessary to pass arguments to the supplied function. An example use case might be the following:

		.. code-block:: python

			def mark_program_end(duration):
			    at(duration)
			    end()

			program = ledsign.LEDSignProgram(device)
			program(mark_program_end, args = (4.0,))
			@program # alternatively: program(program_body_fn)
			def program_body_fn():
			    kp("#ff0000")

		For advanced use cases, command alias injection can be disabled through the :python:`inject_commands` argument (see :py:class:`LEDSignProgramBuilder` for details about command shorthands).
		"""
		if (not callable(func)):
			raise TypeError(f"Expected 'function', got '{func.__class__.__name__}'")
		if (not isinstance(args,tuple) and not isinstance(args,list)):
			raise TypeError(f"Expected 'list' or 'tuple', got '{args.__class__.__name__}'")
		if (not isinstance(kwargs,dict)):
			raise TypeError(f"Expected 'dict', got '{kwargs.__class__.__name__}'")
		self._builder_ready=True
		builder=LEDSignProgramBuilder(self)
		self._builder_ready=False
		builder._change_lock(True)
		if (inject_commands):
			namespace=func.__globals__
			old_namespace={}
			for k,v in builder._get_function_list():
				if (k in namespace):
					old_namespace[k]=namespace[k]
				namespace[k]=v
		try:
			func(*args,**kwargs)
			if (bypass_errors):
				self._has_error=True
			else:
				self.verify()
		except Exception as e:
			self._has_error=True
			raise e
		finally:
			if (inject_commands):
				for k,_ in builder._get_function_list():
					if (k in old_namespace):
						namespace[k]=old_namespace[k]
					else:
						del namespace[k]
			builder._change_lock(False)
		return self

	def _add_raw_keypoint(self,rgb:int,end:int,duration:int,mask:int,frame:object) -> LEDSignKeypoint|None:
		mask&=self._hardware._mask
		if (not mask):
			return None
		out=LEDSignKeypoint(rgb,end,duration,mask,frame)
		self._keypoint_list.insert(out)
		return out

	def _load_from_file(self,file_path:str) -> None:
		size=os.stat(file_path).st_size
		if (size<8 or (size&3)):
			raise LEDSignProgramError("Invalid program")
		with open(file_path,"rb") as rf:
			ctrl,crc=struct.unpack("<II",rf.read(8))
			data=rf.read()
		if ((ctrl&0xff)%3 or size!=((ctrl>>8)<<2)+8 or crc!=LEDSignCRC(data).value):
			raise LEDSignProgramError("Invalid program")
		if (((self._hardware._pixel_count+7)>>3)!=(ctrl&0xff)//3):
			raise LEDSignProgramError("Program was compiled for different hardware")
		self._duration=(ctrl>>9)//max(ctrl&0xff,1)
		parser=LEDSignProgramParser(self,(ctrl&0xff)//3,True)
		parser.update(data)
		parser.terminate()

	def is_unloaded(self) -> bool:
		"""
		Returns :python:`True` if the program is unloaded (ie. not yet fetched from the device), and :python:`False` otherwise.
		"""
		return self._load_parameters is not None

	def get_duration(self) -> float:
		"""
		Returns the current duration of the program.
		"""
		return self._duration/60

	def compile(self,bypass_errors:bool=False) -> LEDSignCompiledProgram:
		"""
		Compiles the program and returns a :py:class:`LEDSignCompiledProgram` object, and optionally bypasses error verification if the :python:`bypass_errors` flag is set. Raises :py:exc:`LEDSignProgramError` if the program contains unresolved errors.
		"""
		self.load()
		if (self._has_error and not bypass_errors):
			raise LEDSignProgramError("Unresolved program errors")
		return LEDSignCompiledProgram(self,False)

	def save(self,file_path:str,bypass_errors:bool=False) -> None:
		"""
		Writes the program to a file pointed to by the :python:`file_path`. Optionally bypasses error verification (if the :python:`bypass_errors` flag is set), or raises :py:exc:`LEDSignProgramError` if the program contains unresolved errors.
		"""
		if (not isinstance(file_path,str)):
			raise TypeError(f"Expected 'str', got '{file_path.__class__.__name__}'")
		self.load()
		if (self._has_error and not bypass_errors):
			raise LEDSignProgramError("Unresolved program errors")
		LEDSignCompiledProgram(self,True)._save_to_file(file_path)

	def load(self) -> None:
		"""
		Explicitly loads an unloaded device program. Does nothing if the program was already downloaded, or if the program was not sourced from a :py:class:`LEDSign` device.

		Raises :py:exc:`LEDSignProtocolError` or :py:exc:`LEDSignProgramError` if the device was closed or modified before this method was called.
		"""
		if (self._load_parameters is None):
			return
		load_parameters=self._load_parameters
		self._load_parameters=None
		device=load_parameters[0]()
		if (device is None or device._path is None):
			self._has_error=True
			raise LEDSignProtocolError("Device disconnected")
		if ((load_parameters[1]&0xff)//3!=self._hardware._led_depth):
			self._has_error=True
			raise LEDSignProgramError("Mismatched program hardware")
		parser=LEDSignProgramParser(self,(load_parameters[1]&0xff)//3,False)
		program_size=(load_parameters[1]>>8)<<2
		chunk_size=min(max(program_size,64),65536)
		chunk_size-=chunk_size%12
		received_crc=LEDSignCRC()
		offset=0
		while (offset<program_size):
			availbale_chunk_size=LEDSignProtocol.process_packet(device._handle,LEDSignProtocol.PACKET_TYPE_PROGRAM_CHUNK_RESPONSE,LEDSignProtocol.PACKET_TYPE_PROGRAM_CHUNK_REQUEST,offset,chunk_size)[0]
			chunk=LEDSignProtocol.process_extended_read(device._handle,availbale_chunk_size)
			received_crc.update(chunk)
			parser.update(chunk)
			offset+=availbale_chunk_size
		if (received_crc.value!=load_parameters[2]):
			self._keypoint_list.clear()
			self._has_error=True
			raise LEDSignProgramError("Mismatched program checksum")
		parser.terminate()

	def get_keypoints(self,mask:int=-1) -> Iterator[LEDSignKeypoint]:
		"""
		Iterates over all keypoints containing any pixels selected by :python:`mask`. If no mask is given, all keypoints are iterated over.
		"""
		if (not isinstance(mask,int)):
			raise TypeError(f"Expected 'int', got '{mask.__class__.__name__}'")
		self.load()
		return self._keypoint_list.iterate(mask)

	def verify(self) -> bool:
		"""
		Verifies the program, and reports any encountered errors. Returns :python:`True` if no errors have been found, and :python:`False` otherwise.
		"""
		self.load()
		self._has_error=False
		kp=self._keypoint_list.lookup_increasing(0,-1)
		while (kp is not None):
			start=kp.end-kp.duration
			if (start<0):
				print(f"Keypoint overlap: ({-start/60:.3f}s)\n  <timeline_start>\n  {kp._frame}",file=LEDSignProgram.error_output_file)
				self._has_error=True
			entry=self._keypoint_list.lookup_decreasing(kp._key-1,kp.mask)
			while (entry is not None and entry.end>start):
				print(f"Keypoint overlap: ({(entry.end-start)/60:.3f}s)\n  {entry._frame}\n  {kp._frame}",file=LEDSignProgram.error_output_file)
				self._has_error=True
				entry=self._keypoint_list.lookup_decreasing(entry._key-1,kp.mask)
			kp=self._keypoint_list.lookup_increasing(kp._key+1,-1)
		return not self._has_error

	@staticmethod
	def _create_unloaded_from_device(device:"LEDSign",ctrl:int,crc:int) -> "LEDSignProgram":
		out=LEDSignProgram(device)
		out._duration=(ctrl>>9)//max(ctrl&0xff,1)
		if (ctrl>>8):
			out._load_parameters=(weakref.ref(device),ctrl,crc)
		return out



class LEDSignProgramBuilder(object):
	"""
	The :py:class:`LEDSignProgramBuilder` class maintains context needed for dynamic program generation. It cannot be created directly; instead it is globally available via :py:func:`LEDSignProgramBuilder.instance` within the stack frames of a decorated :py:class:`LEDSignProgram` function (or function called passed to :py:func:`LEDSignProgram.__call__`).

	For convenience reasons, all program builder commands are made available in the program function's global scope, with shorthand aliases for longer ones. The following commands and shortcuts are injected into the global scope:

	+--------------------+-------------+-------------------------------+
	| Command            | Shortcut    | Builder method                |
	+====================+=============+===============================+
	| :func:`after`      | :func:`af`  | :py:func:`command_after`      |
	+--------------------+-------------+-------------------------------+
	| :func:`at`         | ---         | :py:func:`command_at`         |
	+--------------------+-------------+-------------------------------+
	| :func:`cross_fade` | :func:`cf`  | :py:func:`command_cross_fade` |
	+--------------------+-------------+-------------------------------+
	| :func:`delta_time` | :func:`dt`  | :py:func:`command_delta_time` |
	+--------------------+-------------+-------------------------------+
	| :func:`end`        | ---         | :py:func:`command_end`        |
	+--------------------+-------------+-------------------------------+
	| :func:`hsv`        | ---         | :py:func:`command_hsv`        |
	+--------------------+-------------+-------------------------------+
	| :func:`hardware`   | :func:`hw`  | :py:func:`command_hardware`   |
	+--------------------+-------------+-------------------------------+
	| :func:`keypoint`   | :func:`kp`  | :py:func:`command_keypoint`   |
	+--------------------+-------------+-------------------------------+
	| :func:`rgb`        | ---         | :py:func:`command_rgb`        |
	+--------------------+-------------+-------------------------------+
	| :func:`time`       | :func:`tm`  | :py:func:`command_time`       |
	+--------------------+-------------+-------------------------------+

	The following example illustrates the three different ways of accessing the :python:`keypoint` command:

	.. code-block:: python

		@ledsign.LEDSignProgram(device)
		def program():
		    at(0)
		    keypoint("#ff0000") # <-- alias of the 'keypoint' command
		    after(1)
		    kp("#00ff00") # <-- shortcut for the 'keypoint' command
		    after(1)
		    builder=LEDSignProgramBuilder.instance()
		    builder.command_keypoint("#0000ff") # <-- underlying 'keypoint' command
		    after(1)
		    end()

	.. note::

		Due to limitations of the internal implementation, these shortcuts can only be accessed by code located within the same file (ie. sharing the same global scope) as the executed program generation function. In other contexts (such as in functions where local closures re-define the command aliases), all commands can be accessed through the current program builder instance, ex. :python:`LEDSignProgramBuilder.instance().command_end()`.
	"""

	COMMANDS={
		"af": "after",
		"at": "at",
		"cf": "cross_fade",
		"dt": "delta_time",
		"end": "end",
		"hsv": "hsv",
		"hw": "hardware",
		"kp": "keypoint",
		"rgb": "rgb",
		"tm": "time"
	}

	_global_lock=threading.Lock()
	_current_instance=None

	__slots__=["program","time"]

	def __init__(self,program:LEDSignProgram) -> None:
		if (not isinstance(program,LEDSignProgram) or not program._builder_ready):
			raise TypeError("Direct initialization of LEDSignProgramBuilder is not supported")
		self.program=program
		self.time=1

	def _change_lock(self,enable:bool) -> None:
		if (enable):
			LEDSignProgramBuilder._global_lock.acquire()
			LEDSignProgramBuilder._current_instance=self
		else:
			LEDSignProgramBuilder._current_instance=None
			LEDSignProgramBuilder._global_lock.release()

	def _get_function_list(self) -> Iterator[tuple[str,Callable]]:
		for k,v in LEDSignProgramBuilder.COMMANDS.items():
			func=getattr(self,"command_"+v)
			yield (k,func)
			if (k!=v):
				yield (v,func)

	def _parse_color(self,rgb:int|str) -> int:
		if (isinstance(rgb,int)):
			rgb&=0xffffff
		elif (isinstance(rgb,str) and len(rgb)==7 and rgb[0]=="#"):
			rgb=int(rgb[1:7],16)
		else:
			raise TypeError(f"Expected 'int' or 'hex-color', got '{rgb.__class__.__name__}'")
		return rgb

	def command_at(self,time:int|float) -> float:
		"""
		Sets the current timestamp to :python:`time` (in seconds), and returns this new value. Negative values are clamped to :python:`0.0`.
		"""
		if (not isinstance(time,int) and not isinstance(time,float)):
			raise TypeError(f"Expected 'int' or 'float', got '{time.__class__.__name__}'")
		self.time=max(round(time*60),1)
		return self.time/60

	def command_after(self,delta_time:int|float) -> float:
		"""
		Advances the current timestamp by :python:`delta_time` (in seconds), and returns its new value. Negative values move back the time, clamping to :python:`0.0` if the resulting timestamp becomes negative.
		"""
		if (not isinstance(delta_time,int) and not isinstance(delta_time,float)):
			raise TypeError(f"Expected 'int' or 'float', got '{delta_time.__class__.__name__}'")
		self.time=max(self.time+round(delta_time*60),1)
		return self.time/60

	def command_delta_time(self) -> float:
		"""
		Returns the interval between consecutive program frames, in seconds.
		"""
		return 1/60

	def command_time(self) -> float:
		"""
		Returns the current timestamp, in seconds.
		"""
		return self.time/60

	def command_hardware(self) -> "LEDSignHardware":
		"""
		Returns the :py:class:`LEDSignHardware` object associated with the current program. Can be used as an explicit argument for functions from the :py:class:`LEDSignSelector` class (although these functions will call this method internally if no explicit hardware object is supplied to them).
		"""
		return self.program._hardware

	def command_keypoint(self,rgb:int|str,mask:int=-1,duration:int|float|None=None,time:int|float|None=None) -> LEDSignKeypoint:
		"""
		Creates an :python:`rgb`-colored keypoint. Both integer (:code:`0xrrggbb`) and HTML (:python:`"#rrggbb"`) colors are supported. For other color formats, convert their respective arguments using :py:func:`command_rgb` or :py:func:`command_hsv`.

		If the :python:`mask` argument is given, only pixels selected by the provided mask will have this keypoint applied to them.

		The duration of the animation is specified by the :python:`duration` argument (in seconds). If left undefined or negative, the animation will be instant.

		If the :python:`time` argument is omitted, the keypoint will terminate at the currently selected timestamp. Otherwise, the keypoint will end at timestamp :python:`time`, given in seconds.

		.. warning::

			During program verification, each keypoint is checked to prevent invalid configurations, usually resulting from bugs in generator code. Two types of invalid keypoint configurations are: (1) negative start time, and (2) ambiguous ordering.

			The first error is reported with a keypoint's start time (ie. its end time minus its duration) predates the start of the program (timestamp :python:`0.0`). The compiler is not be able generate the full extent of the animation, and thus the end result is undefined.

			The second error occurs when at any point more than one keypoint attempts to apply an animation to the same pixel. Due to undefined compilation ordering, the produced results may not be consistent across compilations.

			However, for increased performance, these error-checking functions can be postponed or disabled altogether by setting the :python:`bypass_errors` flag in relevant :py:class:`LEDSignProgram` functions.
		"""
		rgb=self._parse_color(rgb)
		if (not isinstance(mask,int)):
			raise TypeError(f"Expected 'int', got '{mask.__class__.__name__}'")
		if (duration is None):
			duration=1
		elif (isinstance(duration,int) or isinstance(duration,float)):
			duration=max(round(duration*60),1)
		else:
			raise TypeError(f"Expected 'int' or 'float', got '{duration.__class__.__name__}'")
		if (time is None):
			time=self.time
		elif (isinstance(time,int) or isinstance(time,float)):
			time=max(round(time*60),1)
		else:
			raise TypeError(f"Expected 'int' or 'float', got '{time.__class__.__name__}'")
		return self.program._add_raw_keypoint(rgb,time,duration,mask,(sys._getframe(1) if hasattr(sys,"_getframe") else None))

	def command_end(self) -> None:
		"""
		Places the program end marker at the current timestamp. All animations after this timestamp will not be compiled. Can be used multiple times to adjust the length of the program.
		"""
		self.program._duration=self.time

	def command_rgb(self,r:int|float,g:int|float,b:int|float) -> int:
		"""
		Converts the given :python:`(r, g, b)` tuple (each element is clamped between :python:`0.0` and :python:`1.0`) into a packed 24-bit integer color. Can be used to pack individual :python:`int` or :python:`float` color channels for use with the :py:func:`command_keypoint` function.
		"""
		if (not isinstance(r,int) and not isinstance(r,float)):
			raise TypeError(f"Expected 'int' or 'float', got '{r.__class__.__name__}'")
		if (not isinstance(g,int) and not isinstance(g,float)):
			raise TypeError(f"Expected 'int' or 'float', got '{g.__class__.__name__}'")
		if (not isinstance(b,int) and not isinstance(b,float)):
			raise TypeError(f"Expected 'int' or 'float', got '{b.__class__.__name__}'")
		r=min(max(round(r*255),0),255)
		g=min(max(round(g*255),0),255)
		b=min(max(round(b*255),0),255)
		return (r<<16)+(g<<8)+b

	def command_hsv(self,h:int|float,s:int|float,v:int|float) -> int:
		"""
		Converts the given :python:`(h, s, v)` tuple (each element is clamped between :python:`0.0` and :python:`1.0`) into a packed 24-bit integer color. Can be used to generate HSV colors for use with the :py:func:`command_keypoint` function.
		"""
		if (not isinstance(h,int) and not isinstance(h,float)):
			raise TypeError(f"Expected 'int' or 'float', got '{h.__class__.__name__}'")
		if (not isinstance(s,int) and not isinstance(s,float)):
			raise TypeError(f"Expected 'int' or 'float', got '{s.__class__.__name__}'")
		if (not isinstance(v,int) and not isinstance(v,float)):
			raise TypeError(f"Expected 'int' or 'float', got '{v.__class__.__name__}'")
		h=(h%1.0)*6
		s=min(max(s,0),1)
		v=min(max(round(v*255),0),255)
		if (s==0):
			return v*0x010101
		i=int(h)
		s*=v
		f=s*(h-i)
		p=min(max(round(v-s),0),255)
		q=min(max(round(v-f),0),255)
		t=min(max(round(v-s+f),0),255)
		if (not i):
			return (v<<16)+(t<<8)+p
		if (i==1):
			return (q<<16)+(v<<8)+p
		if (i==2):
			return (p<<16)+(v<<8)+t
		if (i==3):
			return (p<<16)+(q<<8)+v
		if (i==4):
			return (t<<16)+(p<<8)+v
		return (v<<16)+(p<<8)+q

	def command_cross_fade(self,src:int|str,dst:int|str,t:int|float) -> int:
		"""
		Computes the cross-fade (linearly interpolated) between colors :python:`src` and :python:`dst`, at time :python:`t` (normalized between :python:`0.0` and :python:`1.0`). Both integer (:code:`0xrrggbb`) and HTML (:python:`"#rrggbb"`) colors are supported. For other color formats, convert their respective arguments using :py:func:`command_rgb` or :py:func:`command_hsv`.
		"""
		src=self._parse_color(src)
		dst=self._parse_color(dst)
		if (not isinstance(t,int) and not isinstance(t,float)):
			raise TypeError(f"Expected 'int' or 'float', got '{t.__class__.__name__}'")
		t=min(max(t,0),1)
		cr=min(max(round(((src>>16)&0xff)+t*(((dst>>16)&0xff)-((src>>16)&0xff))),0),255)
		cg=min(max(round(((src>>8)&0xff)+t*(((dst>>8)&0xff)-((src>>8)&0xff))),0),255)
		cb=min(max(round((src&0xff)+t*((dst&0xff)-(src&0xff))),0),255)
		return (cr<<16)+(cg<<8)+cb

	@staticmethod
	def instance() -> Union["LEDSignProgramBuilder",None]:
		"""
		Returns the current active instance of :py:class:`LEDSignProgramBuilder`, or :python:`None` if none are active.
		"""
		return LEDSignProgramBuilder._current_instance



from ledsign.backend import LEDSignProtocolError,LEDSignProtocolBackendWindows,LEDSignProtocolBackendLinux
import struct
import sys



__all__=["LEDSignUnsupportedProtocolError","LEDSignProtocol"]



class LEDSignUnsupportedProtocolError(Exception):
	"""
	Raised when the device responds with an unimplemented protocol version.
	"""



class LEDSignProtocol(object):
	PACKET_TYPE_NONE=0x00
	PACKET_TYPE_HOST_INFO=0x90
	PACKET_TYPE_DEVICE_INFO=0x9f
	PACKET_TYPE_ACK=0xb0
	PACKET_TYPE_LED_DRIVER_STATUS_REQUEST=0x7a
	PACKET_TYPE_LED_DRIVER_STATUS_RESPONSE=0x80
	PACKET_TYPE_PROGRAM_CHUNK_REQUEST=0xd5
	PACKET_TYPE_PROGRAM_CHUNK_REQUEST_DEVICE=0x97
	PACKET_TYPE_PROGRAM_CHUNK_RESPONSE=0xf8
	PACKET_TYPE_PROGRAM_SETUP=0xc8
	PACKET_TYPE_PROGRAM_UPLOAD_STATUS=0xa5
	PACKET_TYPE_HARDWARE_DATA_REQUEST=0x2f
	PACKET_TYPE_HARDWARE_DATA_RESPONSE=0x53

	VERSION=0x0005

	PACKET_FORMATS={
		PACKET_TYPE_NONE: "<BB",
		PACKET_TYPE_HOST_INFO: "<BBH",
		PACKET_TYPE_DEVICE_INFO: "<BBHH8sIIBBBB20sQ",
		PACKET_TYPE_ACK: "<BBB",
		PACKET_TYPE_LED_DRIVER_STATUS_REQUEST: "<BB",
		PACKET_TYPE_LED_DRIVER_STATUS_RESPONSE: "<BBHHII8x",
		PACKET_TYPE_PROGRAM_CHUNK_REQUEST: "<BBII",
		PACKET_TYPE_PROGRAM_CHUNK_REQUEST_DEVICE: "<BBIII",
		PACKET_TYPE_PROGRAM_CHUNK_RESPONSE: "<BBI",
		PACKET_TYPE_PROGRAM_SETUP: "<BBII",
		PACKET_TYPE_PROGRAM_UPLOAD_STATUS: "<BB",
		PACKET_TYPE_HARDWARE_DATA_REQUEST: "<BBB",
		PACKET_TYPE_HARDWARE_DATA_RESPONSE: "<BBHH16x"
	}

	_backend=(LEDSignProtocolBackendWindows if sys.platform=="win32" else LEDSignProtocolBackendLinux)()

	@staticmethod
	def enumerate() -> list[str]:
		return LEDSignProtocol._backend.enumerate()

	@staticmethod
	def open(path:str) -> object:
		return LEDSignProtocol._backend.open(path)

	@staticmethod
	def close(handle:object) -> None:
		LEDSignProtocol._backend.close(handle)

	@staticmethod
	def process_packet(handle:object,ret_type:int,type:int,*args:tuple[int,...]) -> tuple[int,...]:
		ret=LEDSignProtocol._backend.io_read_write(handle,struct.pack(LEDSignProtocol.PACKET_FORMATS[type],type,struct.calcsize(LEDSignProtocol.PACKET_FORMATS[type]),*args))
		if (len(ret)<2 or ret[0]!=ret_type or ret[1]!=len(ret) or ret[1]!=struct.calcsize(LEDSignProtocol.PACKET_FORMATS[ret_type])):
			if (ret_type==LEDSignProtocol.PACKET_TYPE_DEVICE_INFO and type==LEDSignProtocol.PACKET_TYPE_HOST_INFO and len(ret)==2 and ret[0]==LEDSignProtocol.PACKET_TYPE_NONE and ret[1]==2):
				raise LEDSignUnsupportedProtocolError("Protocol version not supported")
			raise LEDSignProtocolError("Protocol error")
		return struct.unpack(LEDSignProtocol.PACKET_FORMATS[ret_type],ret)[2:]

	@staticmethod
	def process_extended_read(handle:object,size:int) -> bytearray:
		return LEDSignProtocol._backend.io_bulk_read(handle,size)

	@staticmethod
	def process_extended_write(handle:object,data:bytearray) -> int:
		return LEDSignProtocol._backend.io_bulk_write(handle,data)



from collections.abc import Callable
from ledsign.hardware import LEDSignHardware
from ledsign.program import LEDSignProgram
from ledsign.program_io import LEDSignCompiledProgram
from ledsign.protocol import LEDSignProtocol
import time



__all__=["LEDSignDeviceNotFoundError","LEDSignAccessError","LEDSign"]



class LEDSignDeviceNotFoundError(Exception):
	"""
	Raised by :py:func:`LEDSign.open` when no device was found.
	"""



class LEDSignAccessError(Exception):
	"""
	Raised when the device access mode was violated.
	"""



class LEDSign(object):
	"""
	Returned by :py:func:`LEDSign.open`, represents a handle to an LED sign device.

	.. autoattribute:: ACCESS_MODE_NONE
	   :no-value:

	   Access mode representing a device handle without read or write access. At the moment only used by closed device handles.

	.. autoattribute:: ACCESS_MODE_READ
	   :no-value:

	   Access mode representing a device handle with only read permissions.

	.. autoattribute:: ACCESS_MODE_READ_WRITE
	   :no-value:

	   Access mode representing a device handle with both read and write permissions.
	"""

	ACCESS_MODE_NONE:int=0x00
	ACCESS_MODE_READ:int=0x01
	ACCESS_MODE_READ_WRITE:int=0x02

	ACCESS_MODES:dict[int,str]={
		ACCESS_MODE_NONE: "none",
		ACCESS_MODE_READ: "read-only",
		ACCESS_MODE_READ_WRITE: "read-write",
	}

	__slots__=["__weakref__","_path","_handle","_access_mode","_psu_current","_storage_size","_hardware","_firmware","_serial_number","_driver_brightness","_driver_program_paused","_driver_temperature","_driver_load","_driver_program_time","_driver_current_usage","_driver_program_offset_divisor","_driver_info_sync_next_time","_driver_info_sync_interval","_program"]

	def __init__(self,path,handle,config_packet) -> None:
		self._path=path
		self._handle=handle
		self._access_mode=config_packet[6]&0x0f
		self._psu_current=(config_packet[7]&0x7f)/10
		self._storage_size=config_packet[1]<<10
		self._hardware=LEDSignHardware(handle,config_packet[2])
		self._firmware=config_packet[9].hex()
		self._serial_number=config_packet[10]
		self._driver_brightness=config_packet[5]&0x0f
		self._driver_program_paused=not (config_packet[8]&1)
		self._driver_program_offset_divisor=max((config_packet[3]&0xff)<<1,1)*60
		self._driver_info_sync_next_time=0
		self._driver_info_sync_interval=0.5
		self._program=LEDSignProgram._create_unloaded_from_device(self,config_packet[3],config_packet[4])

	def __del__(self) -> None:
		if (self._handle is not None):
			LEDSignProtocol.close(self._handle)

	def __repr__(self) -> str:
		return f"<LEDSign id={self._serial_number:016x} fw={self._firmware}>"

	def _check_if_closed(self) -> None:
		if (self._handle is not None):
			return
		raise LEDSignAccessError("Device handle was closed, no access allowed")

	def _sync_driver_info(self) -> None:
		self._check_if_closed()
		if (time.time()<self._driver_info_sync_next_time):
			return
		driver_status=LEDSignProtocol.process_packet(self._handle,LEDSignProtocol.PACKET_TYPE_LED_DRIVER_STATUS_RESPONSE,LEDSignProtocol.PACKET_TYPE_LED_DRIVER_STATUS_REQUEST)
		self._driver_temperature=437.226612-driver_status[0]*0.468137
		self._driver_load=driver_status[1]/160
		self._driver_program_time=driver_status[2]/self._driver_program_offset_divisor
		self._driver_current_usage=driver_status[3]*1e-6
		self._driver_info_sync_next_time=time.time()+self._driver_info_sync_interval

	def close(self) -> None:
		"""
		Closes the underlying device handle. After a call to this function, all methods except for :py:func:`get_path` and :py:func:`get_access_mode` will raise a :py:exc:`LEDSignAccessError` exception.
		"""
		self._check_if_closed()
		LEDSignProtocol.close(self._handle)
		self._handle=None
		self._path=None
		self._access_mode=LEDSign.ACCESS_MODE_NONE

	def get_path(self) -> str|None:
		"""
		Returns the underlying OS path of the device, or :python:`None` if the device was closed.
		"""
		return self._path

	def get_access_mode(self) -> int:
		"""
		Returns the access mode (permissions) granted by the device. Possible return values are:

		* :py:attr:`ACCESS_MODE_NONE`: No access; device handle was closed
		* :py:attr:`ACCESS_MODE_READ`: Read-only access; program uploads will be rejected
		* :py:attr:`ACCESS_MODE_READ_WRITE`: Full read-write access
		"""
		return self._access_mode

	def get_access_mode_str(self) -> str:
		"""
		Same as :py:func:`get_access_mode`, but returns a stringified versions of the access mode. Possible values are: :python:`"none"`, :python:`"read-only"`, or :python:`"read-write"`.
		"""
		return LEDSign.ACCESS_MODES[self._access_mode]

	def get_psu_current(self) -> float:
		"""
		Returns the configured theoretical current limit of the power supply, in amps. As only 5V power supplies are supported, no explicit voltage getter method is provided.

		.. danger::
		   If the device draws more current than this limit, a device-internal overcurrent safety flag will be raised. Whenever this flag is active, no changes made to the device will be visible.

		   **For safety reasons, this flag can only be cleared from the UI menu, or through device reboots.**
		"""
		self._check_if_closed()
		return self._psu_current

	def get_storage_size(self) -> int:
		"""
		Returns the storage capacity of the device allocated for program storage, in bytes.
		"""
		self._check_if_closed()
		return self._storage_size

	def get_hardware(self) -> LEDSignHardware:
		"""
		Returns a read-only :py:class:`LEDSignHardware` object representing the current external hardware configuration of the device.
		"""
		self._check_if_closed()
		return self._hardware

	def get_firmware(self) -> str:
		"""
		Returns a 40-character hex string uniquely identifying the current firmware versions of the device.
		"""
		self._check_if_closed()
		return self._firmware

	def get_serial_number(self) -> int:
		"""
		Returns the unique 64-bit serial number of the device.
		"""
		self._check_if_closed()
		return self._serial_number

	def get_serial_number_str(self) -> str:
		"""
		Same as :py:func:`get_serial_number`, but formats the serial number as a 16-hex-digit string.
		"""
		self._check_if_closed()
		return f"{self._serial_number:016x}"

	def get_driver_brightness(self) -> float:
		"""
		Returns the current device brightness setting, normalized between :python:`0.0` and :python:`1.0`.

		.. note::
		   If the overcurrent flag was tripped (see :py:func:`get_psu_current` for details), the returned brightness setting may not reflect the real-world conditions.
		"""
		self._check_if_closed()
		return round(self._driver_brightness*20/7)/20

	def is_driver_paused(self) -> bool:
		"""
		Returns :python:`True` if the LED driver is in a paused state (ie. the program playback is frozen), and :python:`False` otherwise.

		.. note::
		   For internal protocol reasons, this flag is **not** periodically synchronized with the device.
		"""
		self._check_if_closed()
		return self._driver_program_paused

	def get_driver_temperature(self) -> float:
		"""
		Returns the current temperature of the LED driver, in degrees Celsius.

		The temperature is fetched periodically from the device, at an interval specified by :py:func:`get_driver_status_reload_time`.
		"""
		self._sync_driver_info()
		return self._driver_temperature

	def get_driver_load(self) -> float:
		"""
		Returns the current driver CPU load, normalized between :python:`0.0` and :python:`1.0`. The returned value represents the frame render duration, measured as a fraction of the overall frame time.

		The driver load is fetched periodically from the device, at an interval specified by :py:func:`get_driver_status_reload_time`.
		"""
		self._sync_driver_info()
		return self._driver_load

	def get_driver_program_time(self) -> float:
		"""
		Returns the current program timestamp. The timestamp wraps around to zero at program duration, ie. :python:`timestamp %= self.get_program().get_duration()`.

		The timestamp is fetched periodically from the device, at an interval specified by :py:func:`get_driver_status_reload_time`.
		"""
		self._sync_driver_info()
		return self._driver_program_time

	def get_driver_current_usage(self) -> float:
		"""
		Returns the current being drawn by the device hardware from the power supply. Internally, this same value is used in the overcurrent protection circuit (see :py:func:`get_psu_current` for more details).

		The current usage is fetched periodically from the device, at an interval specified by :py:func:`get_driver_status_reload_time`.
		"""
		self._sync_driver_info()
		return self._driver_current_usage

	def get_driver_status_reload_time(self) -> float:
		"""
		Returns the current refresh interval (in seconds) used for fetching driver data, or :python:`-1` if caching is disabled. Can be modified using :py:func:`set_driver_status_reload_time`.
		"""
		self._check_if_closed()
		return self._driver_info_sync_interval

	def set_driver_status_reload_time(self,interval:float) -> float:
		"""
		Sets :python:`interval` as the new driver data refresh interval (in seconds), and returns the previous value. Negative values for :python:`interval` disable caching and fetch data whenever it is requested.
		"""
		self._check_if_closed()
		out=self._driver_info_sync_interval
		self._driver_info_sync_interval=(-1 if interval<=0 else interval)
		self._driver_info_sync_next_time+=self._driver_info_sync_interval-out
		return out

	def get_program(self) -> LEDSignProgram:
		"""
		Returns a read-only :py:class:`LEDSignProgram` object containing the current device program. For performance reasons, the program is lazily loaded only when its keypoints are accessed to prevent significant start-up delays.
		"""
		self._check_if_closed()
		return self._program

	def upload_program(self,program:LEDSignCompiledProgram,callback:Callable[[float,bool],None]|None=None) -> None:
		"""
		Uploads the given :python:`program` to the device. If the device was opened in read-only mode, a :py:exc:`LEDSignAccessError` exception will be raised.

		The optional :python:`callback` argument is periodically called throughout the upload process with arguments :python:`(progress, is_first_stage)`.
		"""
		if (not isinstance(program,LEDSignCompiledProgram)):
			raise TypeError(f"Expected 'LEDSignCompiledProgram', got '{program.__class__.__name__}'")
		if (callback is not None and not callable(callback)):
			raise TypeError(f"Expected 'function', got '{callback.__class__.__name__}'")
		if (self._access_mode!=LEDSign.ACCESS_MODE_READ_WRITE):
			raise LEDSignAccessError("Program upload not allowed, Python API configured as read-only")
		program._upload_to_device(self,callback)

	@staticmethod
	def open(path:str|None=None) -> "LEDSign":
		"""
		Opens the device pointed to by :python:`path` (or the default device if no path is provided), and returns its corresponding :py:class:`LEDSign` object. The device path must have been returned by a previous call to :py:func:`enumerate`.

		If the device is already in use by a different program, a :py:exc:`LEDSignDeviceInUseError` will be raised. Additionally, if the device protocol version is incompatible, or if the Python API was disabled in device settings, a :py:exc:`LEDSignUnsupportedProtocolError` or :py:exc:`LEDSignProtocolError` exception will be returned respectively.

		.. note::
		   Due to the internal USB-level configuration, this library can be used alongside the UI without triggering a :py:exc:`LEDSignDeviceInUseError` exception. This feature allows the UI to be kept open while writing new programs to ease the debugging process.
		"""
		if (path is None):
			devices=LEDSignProtocol.enumerate()
			if (not devices):
				raise LEDSignDeviceNotFoundError("No device found")
			path=devices[0]
		if (not isinstance(path,str)):
			raise TypeError(f"Expected 'str', got '{path.__class__.__name__}'")
		handle=None
		try:
			handle=LEDSignProtocol.open(path)
			config_packet=LEDSignProtocol.process_packet(handle,LEDSignProtocol.PACKET_TYPE_DEVICE_INFO,LEDSignProtocol.PACKET_TYPE_HOST_INFO,LEDSignProtocol.VERSION)
		except Exception as e:
			if (handle is not None):
				LEDSignProtocol.close(handle)
			raise e
		return LEDSign(path,handle,config_packet)

	@staticmethod
	def enumerate() -> list[str]:
		"""
		Returns a list of available device paths detected by the OS.
		"""
		return LEDSignProtocol.enumerate()



import ctypes
import os
import struct



__all__=["LEDSignProtocolError","LEDSignDeviceInUseError","LEDSignProtocolBackendWindows","LEDSignProtocolBackendLinux"]



class LEDSignProtocolError(Exception):
	"""
	Raised during any encountered protocol error.
	"""



class LEDSignDeviceInUseError(Exception):
	"""
	Raised when a device path supplied to :py:func:`LEDSign.open` was already in use by a different program.
	"""



class LEDSignProtocolBackendWindows(object):
	CM_GET_DEVICE_INTERFACE_LIST_ALL_DEVICES=0x00000000
	CR_BUFFER_SMALL=0x0000001a
	GENERIC_WRITE=0x40000000
	GENERIC_READ=0x80000000
	FILE_SHARE_READ=0x00000001
	FILE_SHARE_WRITE=0x00000002
	OPEN_EXISTING=0x00000003
	FILE_ATTRIBUTE_NORMAL=0x00000080
	FILE_FLAG_OVERLAPPED=0x40000000

	def __init__(self) -> None:
		import ctypes.wintypes
		GUID=type("GUID",(ctypes.Structure,),{"_fields_":[("Data1",ctypes.wintypes.DWORD),("Data2",ctypes.wintypes.WORD),("Data3",ctypes.wintypes.WORD),("Data4",ctypes.wintypes.BYTE*8)]})
		PGUID=ctypes.POINTER(GUID)
		self.CM_Get_Device_Interface_List_SizeA=ctypes.windll.cfgmgr32.CM_Get_Device_Interface_List_SizeA
		self.CM_Get_Device_Interface_List_SizeA.argtypes=(ctypes.wintypes.PULONG,PGUID,ctypes.wintypes.LPVOID,ctypes.wintypes.ULONG)
		self.CM_Get_Device_Interface_List_SizeA.restype=ctypes.wintypes.DWORD
		self.CM_Get_Device_Interface_ListA=ctypes.windll.cfgmgr32.CM_Get_Device_Interface_ListA
		self.CM_Get_Device_Interface_ListA.argtypes=(PGUID,ctypes.wintypes.LPVOID,ctypes.wintypes.PCHAR,ctypes.wintypes.ULONG,ctypes.wintypes.ULONG)
		self.CM_Get_Device_Interface_ListA.restype=ctypes.wintypes.DWORD
		self.CreateFileA=ctypes.windll.kernel32.CreateFileA
		self.CreateFileA.argtypes=(ctypes.wintypes.LPCSTR,ctypes.wintypes.DWORD,ctypes.wintypes.DWORD,ctypes.wintypes.LPVOID,ctypes.wintypes.DWORD,ctypes.wintypes.DWORD,ctypes.wintypes.HANDLE)
		self.CreateFileA.restype=ctypes.wintypes.HANDLE
		self.CloseHandle=ctypes.windll.kernel32.CloseHandle
		self.CloseHandle.argtypes=(ctypes.wintypes.HANDLE,)
		self.CloseHandle.restype=ctypes.wintypes.BOOL
		self.WinUsb_Initialize=ctypes.windll.winusb.WinUsb_Initialize
		self.WinUsb_Initialize.argtypes=(ctypes.wintypes.HANDLE,ctypes.POINTER(ctypes.wintypes.HANDLE))
		self.WinUsb_Initialize.restype=ctypes.wintypes.BOOL
		self.WinUsb_Free=ctypes.windll.winusb.WinUsb_Free
		self.WinUsb_Free.argtypes=(ctypes.wintypes.HANDLE,)
		self.WinUsb_Free.restype=ctypes.wintypes.BOOL
		self.WinUsb_ControlTransfer=ctypes.windll.winusb.WinUsb_ControlTransfer
		self.WinUsb_ControlTransfer.argtypes=(ctypes.wintypes.HANDLE,ctypes.c_ulonglong,ctypes.wintypes.PCHAR,ctypes.wintypes.ULONG,ctypes.wintypes.PULONG,ctypes.wintypes.LPVOID)
		self.WinUsb_ControlTransfer.restype=ctypes.wintypes.BOOL
		self.WinUsb_ReadPipe=ctypes.windll.winusb.WinUsb_ReadPipe
		self.WinUsb_ReadPipe.argtypes=(ctypes.wintypes.HANDLE,ctypes.wintypes.CHAR,ctypes.wintypes.PCHAR,ctypes.wintypes.ULONG,ctypes.wintypes.PULONG,ctypes.wintypes.LPVOID)
		self.WinUsb_ReadPipe.restype=ctypes.wintypes.BOOL
		self.WinUsb_WritePipe=ctypes.windll.winusb.WinUsb_WritePipe
		self.WinUsb_WritePipe.argtypes=(ctypes.wintypes.HANDLE,ctypes.wintypes.CHAR,ctypes.wintypes.PCHAR,ctypes.wintypes.ULONG,ctypes.wintypes.PULONG,ctypes.wintypes.LPVOID)
		self.WinUsb_WritePipe.restype=ctypes.wintypes.BOOL
		self.winusb_registry_guid_ref=ctypes.byref(GUID.from_buffer(bytearray(b"\x58\x30\xf5\xfc\x7b\x99\x21\x4b\xaf\xd6\xe5\x65\x04\x39\x23\xc1")))

	def enumerate(self) -> list[str]:
		while (True):
			length=ctypes.wintypes.ULONG(0)
			if (self.CM_Get_Device_Interface_List_SizeA(ctypes.byref(length),self.winusb_registry_guid_ref,0,LEDSignProtocolBackendWindows.CM_GET_DEVICE_INTERFACE_LIST_ALL_DEVICES)):
				raise OSError("CM_Get_Device_Interface_List_SizeA error")
			data=(ctypes.wintypes.CHAR*length.value)()
			ret=self.CM_Get_Device_Interface_ListA(self.winusb_registry_guid_ref,0,data,length,LEDSignProtocolBackendWindows.CM_GET_DEVICE_INTERFACE_LIST_ALL_DEVICES)
			if (not ret):
				return [e for e in bytes(data).decode("utf-8").split("\x00") if e]
			if (ret==LEDSignProtocolBackendWindows.CR_BUFFER_SMALL):
				continue
			raise OSError("CM_Get_Device_Interface_ListA error")

	def open(self,path:str) -> tuple[int,int]:
		handle=self.CreateFileA(path.encode("utf-8"),LEDSignProtocolBackendWindows.GENERIC_WRITE|LEDSignProtocolBackendWindows.GENERIC_READ,LEDSignProtocolBackendWindows.FILE_SHARE_READ|LEDSignProtocolBackendWindows.FILE_SHARE_WRITE,0,LEDSignProtocolBackendWindows.OPEN_EXISTING,LEDSignProtocolBackendWindows.FILE_ATTRIBUTE_NORMAL|LEDSignProtocolBackendWindows.FILE_FLAG_OVERLAPPED,0)
		if (handle==0xffffffffffffffff):
			raise LEDSignDeviceInUseError("Device already in use")
		winusb_handle=ctypes.wintypes.HANDLE(0)
		if (not self.WinUsb_Initialize(handle,ctypes.byref(winusb_handle))):
			self.CloseHandle(handle)
			raise LEDSignDeviceInUseError("Device already in use")
		buffer=(ctypes.wintypes.CHAR*64)()
		transferred=ctypes.c_ulong(0)
		if (not self.WinUsb_ControlTransfer(winusb_handle.value,0x00400000545352c0,buffer,64,ctypes.byref(transferred),0) or transferred.value!=5 or bytes(buffer[:5])!=b"reset"):
			self.WinUsb_Free(winusb_handle.value)
			self.CloseHandle(handle)
			raise LEDSignProtocolError("Unable to reset device, Python API disabled")
		return (handle,winusb_handle.value)

	def close(self,handles:tuple[int,int]) -> None:
		self.WinUsb_Free(handles[1])
		self.CloseHandle(handles[0])

	def io_read_write(self,handles:tuple[int,int],packet:bytes) -> bytearray:
		packet=bytearray(packet)
		transferred=ctypes.c_ulong(0)
		if (not self.WinUsb_WritePipe(handles[1],0x04,(ctypes.c_char*len(packet)).from_buffer(packet),len(packet),ctypes.byref(transferred),0) or transferred.value!=len(packet)):
			raise LEDSignProtocolError("Write to endpoint 04h failed")
		transferred.value=0
		out=(ctypes.c_char*64)()
		if (not self.WinUsb_ReadPipe(handles[1],0x84,out,64,ctypes.byref(transferred),0) or transferred.value<2 or transferred.value>64):
			raise LEDSignProtocolError("Read from endpoint 84h failed")
		return bytearray(out)[:transferred.value]

	def io_bulk_read(self,handles:tuple[int,int],size:int) -> bytearray:
		out=(ctypes.c_char*size)()
		transferred=ctypes.c_ulong(0)
		if (not self.WinUsb_ReadPipe(handles[1],0x85,out,size,ctypes.byref(transferred),0) or transferred.value!=size):
			raise LEDSignProtocolError("Read from endpoint 85h failed")
		return bytearray(out)

	def io_bulk_write(self,handles:tuple[int,int],data) -> None:
		data=bytearray(data)
		transferred=ctypes.c_ulong(0)
		if (not self.WinUsb_WritePipe(handles[1],0x05,(ctypes.c_char*len(data)).from_buffer(data),len(data),ctypes.byref(transferred),0) or transferred.value!=len(data)):
			raise LEDSignProtocolError("Write to endpoint 05h failed")



class LEDSignProtocolBackendLinux(object):
	USBDEVFS_BULK=0xc0185502
	USBDEVFS_CLAIMINTERFACE=0x8004550f
	USBDEVFS_CONTROL=0xc0185500

	def __init__(self) -> None:
		import ctypes.util
		self.ioctl=ctypes.CDLL(ctypes.util.find_library("c"),use_errno=True).ioctl
		self.ioctl.argtypes=(ctypes.c_int,ctypes.c_ulong,ctypes.c_char_p)
		self.ioctl.restype=ctypes.c_int

	def enumerate(self) -> list[str]:
		out=[]
		for name in os.listdir("/sys/bus/usb/devices"):
			if (not os.path.exists(f"/sys/bus/usb/devices/{name}/idVendor")):
				continue
			with open(f"/sys/bus/usb/devices/{name}/idVendor","rb") as rf:
				if (rf.read()!=b"fff0\n"):
					continue
			with open(f"/sys/bus/usb/devices/{name}/idProduct","rb") as rf:
				if (rf.read()!=b"1000\n"):
					continue
			with open(f"/sys/bus/usb/devices/{name}/busnum","rb") as rf:
				busnum=int(rf.read())
			with open(f"/sys/bus/usb/devices/{name}/devnum","rb") as rf:
				devnum=int(rf.read())
			out.append(f"/dev/bus/usb/{busnum:03d}/{devnum:03d}")
		return out

	def open(self,path:str) -> int:
		handle=os.open(path,os.O_RDWR)
		if (self.ioctl(handle,LEDSignProtocolBackendLinux.USBDEVFS_CLAIMINTERFACE,struct.pack("<I",1))<0):
			raise LEDSignDeviceInUseError("Device already in use")
		buffer=(ctypes.c_uint8*64)()
		if (self.ioctl(handle,LEDSignProtocolBackendLinux.USBDEVFS_CONTROL,struct.pack("<BBHHHI4xQ",0xc0,0x52,0x5453,0,64,3000,ctypes.addressof(buffer)))!=5 or bytes(buffer[:5])!=b"reset"):
			os.close(handle)
			raise LEDSignProtocolError("Unable to reset device, Python API disabled")
		return handle

	def close(self,handle:int) -> None:
		os.close(handle)

	def io_read_write(self,handle:int,packet:bytes) -> bytearray:
		packet=bytearray(packet)
		if (self.ioctl(handle,LEDSignProtocolBackendLinux.USBDEVFS_BULK,struct.pack("<III4xQ",0x04,len(packet),3000,ctypes.addressof((ctypes.c_uint8*len(packet)).from_buffer(packet))))!=len(packet)):
			raise LEDSignProtocolError("Write to endpoint 04h failed")
		out=(ctypes.c_uint8*64)()
		ret=self.ioctl(handle,LEDSignProtocolBackendLinux.USBDEVFS_BULK,struct.pack("<III4xQ",0x84,64,3000,ctypes.addressof(out)))
		if (ret<2 or ret>64):
			raise LEDSignProtocolError("Read from endpoint 84h failed")
		return bytearray(out)[:ret]

	def io_bulk_read(self,handle:int,size:int) -> bytearray:
		out=(ctypes.c_uint8*size)()
		ret=self.ioctl(handle,LEDSignProtocolBackendLinux.USBDEVFS_BULK,struct.pack("<III4xQ",0x85,size,3000,ctypes.addressof(out)))
		if (ret!=size):
			raise LEDSignProtocolError("Read from endpoint 85h failed")
		return bytearray(out)

	def io_bulk_write(self,handle:int,data:bytearray) -> None:
		if (self.ioctl(handle,LEDSignProtocolBackendLinux.USBDEVFS_BULK,struct.pack("<III4xQ",0x05,len(data),3000,ctypes.addressof((ctypes.c_uint8*len(data)).from_buffer(data))))!=len(data)):
			raise LEDSignProtocolError("Write to endpoint 05h failed")



from collections.abc import Iterator
from ledsign.program import LEDSignProgramBuilder
from ledsign.protocol import LEDSignProtocol
import array



__all__=["LEDSignHardware","LEDSignSelector"]



class LEDSignHardware(object):
	"""
	Read-only representation of a :py:class:`LEDSign` device's hardware, returned by :py:func:`LEDSign.get_hardware()`.
	"""

	SCALE=1/768

	__slots__=["_raw_config","_led_depth","_pixels","_pixel_count","_max_x","_max_y","_mask"]

	def __init__(self,handle,config) -> None:
		if (not isinstance(config,bytes) or len(config)!=8):
			raise TypeError("Direct initialization of LEDSignHardware is not supported")
		self._raw_config=config
		self._led_depth=0
		self._pixels=[]
		self._pixel_count=0
		self._max_x=0
		self._max_y=0
		self._mask=0
		width_map={0:0}
		geometry_map={0:array.array("H")}
		for i in range(0,8):
			key=self._raw_config[i]
			if (key not in geometry_map):
				geometry_map[key]=array.array("I")
				length,width=LEDSignProtocol.process_packet(handle,LEDSignProtocol.PACKET_TYPE_HARDWARE_DATA_RESPONSE,LEDSignProtocol.PACKET_TYPE_HARDWARE_DATA_REQUEST,key)
				width_map[key]=width*LEDSignHardware.SCALE
				geometry_map[key].frombytes(LEDSignProtocol.process_extended_read(handle,length))
			self._led_depth=max(self._led_depth,len(geometry_map[key]))
		for i in range(0,8):
			geometry=geometry_map.get(self._raw_config[i],[])
			self._pixel_count+=len(geometry)
			for xy in geometry:
				x=(xy&0xffff)*LEDSignHardware.SCALE
				y=(xy>>16)*LEDSignHardware.SCALE
				self._max_y=max(self._max_y,y)
				self._mask|=1<<len(self._pixels)
				self._pixels.append((self._max_x+x,y))
			self._max_x+=width_map[self._raw_config[i]]
			for j in range(len(geometry),self._led_depth):
				self._pixels.append(None)
		self._max_x=max(self._max_x,0)

	def __repr__(self) -> str:
		return f"<LEDSignHardware config={self.get_string()} pixels={self._pixel_count}>"

	def get_raw(self) -> bytes:
		"""
		Returns the raw, 8-byte hardware configuration sequence returned by the device.
		"""
		return self._raw_config

	def get_string(self) -> str:
		"""
		Returns a pretty-printed version of the raw device hardware configuration returned by :py:func:`get_raw`.
		"""
		return "["+" ".join([f"{e:02x}" for e in self._raw_config])+"]"

	def get_user_string(self) -> str:
		"""
		Returns a user-friendly string identifying the device's hardware configuration.
		"""
		return bytearray([e for e in self._raw_config if e]).decode("utf-8")



class LEDSignSelector(object):
	"""
	This static class contains several common pattern generators, to be used within a :py:class:`LEDSignProgramBuilder` context.

	Every method accepts an optional :python:`hardware` argument, which when omitted will default to the one used by the current :py:class:`LEDSignProgramBuilder` instance (if a program builder is currently active).
	"""
	@staticmethod
	def _check_hardware(hardware:LEDSignHardware|None) -> LEDSignHardware:
		if (hardware is None):
			builder=LEDSignProgramBuilder.instance()
			if (builder is not None):
				hardware=builder.program._hardware
		if (not isinstance(hardware,LEDSignHardware)):
			raise TypeError(f"Expected 'LEDSignHardware', got '{hardware.__class__.__name__}'")
		return hardware

	@staticmethod
	def get_mask(hardware:LEDSignHardware|None=None) -> int:
		"""
		Returns a mask selecting all pixels.
		"""
		return LEDSignSelector._check_hardware(hardware)._mask

	@staticmethod
	def get_led_depth(hardware:LEDSignHardware|None=None) -> int:
		"""
		Returns the longest LED chain present in the current hardware configuration.
		"""
		return LEDSignSelector._check_hardware(hardware)._led_depth

	@staticmethod
	def get_bounding_box(mask:int=-1,hardware:LEDSignHardware|None=None) -> tuple[float,float,float,float]:
		"""
		Returns the bounding box :python:`(sx, sy, ex, ey)` of all pixels selected by :python:`mask`. If no mask is given, the bounding box of all pixels is computed.
		"""
		hardware=LEDSignSelector._check_hardware(hardware)
		if (not isinstance(mask,int)):
			raise TypeError(f"Expected 'int', got '{mask.__class__.__name__}'")
		out=[0.0,0.0,0.0,0.0]
		is_first=True
		for i,xy in enumerate(hardware._pixels):
			if (xy is not None and (mask&1)):
				x,y=xy
				if (is_first):
					is_first=False
					out[0]=x
					out[1]=y
					out[2]=x
					out[3]=y
				else:
					out[0]=min(out[0],x)
					out[1]=min(out[1],y)
					out[2]=max(out[2],x)
					out[3]=max(out[3],y)
			mask>>=1
		return tuple(out)

	@staticmethod
	def get_center(mask:int=-1,weighted:bool=False,hardware:LEDSignHardware|None=None) -> tuple[float,float]:
		"""
		Returns the center :python:`(cx, cy)` of all pixels selected by :python:`mask`. If no mask is given, the bounding box of all pixels is computed.

		If the :python:`weighted` flag is set, the center is weighted across all pixels locations. Otherwise, the center of the bounding box returned by :py:func:`get_bounding_box` is calculated.
		"""
		hardware=LEDSignSelector._check_hardware(hardware)
		if (not isinstance(mask,int)):
			raise TypeError(f"Expected 'int', got '{mask.__class__.__name__}'")
		if (not weighted):
			bbox=LEDSignSelector.get_bounding_box(mask,hardware)
			return ((bbox[0]+bbox[2])/2,(bbox[1]+bbox[3])/2)
		cx=0.0
		cy=0.0
		cn=0
		for i,xy in enumerate(hardware._pixels):
			if (xy is not None and (mask&1)):
				cx+=xy[0]
				cy+=xy[1]
				cn+=1
			mask>>=1
		cn+=not cn
		return (cx/cn,cy/cn)

	@staticmethod
	def get_pixels(mask:int=-1,letter:int|None=None,hardware:LEDSignHardware|None=None) -> Iterator[tuple[float,float,int]]:
		"""
		Returns all pixel :python:`(x, y, mask)` tuples selected by :python:`mask`. If no mask is given, the bounding box of all pixels is computed.

		If the :python:`letter` index is given, only pixels from the given letter are processed.
		"""
		hardware=LEDSignSelector._check_hardware(hardware)
		if (not isinstance(mask,int)):
			raise TypeError(f"Expected 'int', got '{mask.__class__.__name__}'")
		if (letter is not None):
			mask&=LEDSignSelector.get_letter_mask(letter,hardware=hardware)
		m=1
		for i,xy in enumerate(hardware._pixels):
			if (xy is not None and (mask&m)):
				yield (xy[0],xy[1],m)
			m<<=1

	@staticmethod
	def get_letter_mask(index:int,hardware:LEDSignHardware|None=None) -> int:
		"""
		Returns the pixel mask corresponding to the letter selected by :python:`index`. Raises an :py:exc:`IndexError` if the index is out of bounds.
		"""
		hardware=LEDSignSelector._check_hardware(hardware)
		if (not isinstance(index,int)):
			raise TypeError(f"Expected 'int', got '{index.__class__.__name__}'")
		if (index<0):
			raise IndexError("Letter index out of range")
		for i in range(0,8):
			if (not hardware._raw_config[i]):
				continue
			if (not index):
				return ((1<<((i+1)*hardware._led_depth))-(1<<(i*hardware._led_depth)))&hardware._mask
			index-=1
		raise IndexError("Letter index out of range")

	@staticmethod
	def get_letter_masks(hardware:LEDSignHardware|None=None) -> Iterator[tuple[int,str,int]]:
		"""
		Returns all letter :python:`(index, character, mask)` tuples of the current hardware configuration.
		"""
		hardware=LEDSignSelector._check_hardware(hardware)
		j=0
		for i in range(0,8):
			if (not hardware._raw_config[i]):
				continue
			yield (j,chr(hardware._raw_config[i]),((1<<((i+1)*hardware._led_depth))-(1<<(i*hardware._led_depth)))&hardware._mask)
			j+=1

	@staticmethod
	def get_letter_count(hardware:LEDSignHardware|None=None) -> int:
		"""
		Returns the number of letters in the current hardware configuration.
		"""
		hardware=LEDSignSelector._check_hardware(hardware)
		out=0
		for i in range(0,8):
			if (hardware._raw_config[i]):
				out+=1
		return out

	@staticmethod
	def get_circle_mask(cx:int|float,cy:int|float,r:int|float,mask:int=-1,hardware:LEDSignHardware|None=None) -> int:
		"""
		Returns a mask selecting all pixels within the circle centered at :python:`(cx, cy)` with radius :python:`r`, selected by the provided :python:`mask`. If no mask is given, the all pixels are processed.
		"""
		hardware=LEDSignSelector._check_hardware(hardware)
		if (not isinstance(cx,int) and not isinstance(cx,float)):
			raise TypeError(f"Expected 'int' or 'float', got '{cx.__class__.__name__}'")
		if (not isinstance(cy,int) and not isinstance(cy,float)):
			raise TypeError(f"Expected 'int' or 'float', got '{cy.__class__.__name__}'")
		if (not isinstance(r,int) and not isinstance(r,float)):
			raise TypeError(f"Expected 'int' or 'float', got '{r.__class__.__name__}'")
		if (r<0):
			raise ValueError(f"Radius must not be negative, got '{r}'")
		if (not isinstance(mask,int)):
			raise TypeError(f"Expected 'int', got '{mask.__class__.__name__}'")
		r*=r
		out=0
		for i,xy in enumerate(hardware._pixels):
			if (xy is not None and (mask&1) and (xy[0]-cx)**2+(xy[1]-cy)**2<=r):
				out|=1<<i
			mask>>=1
		return out



from ledsign.backend import LEDSignProtocolError,LEDSignDeviceInUseError
from ledsign.device import LEDSignDeviceNotFoundError,LEDSignAccessError,LEDSign
from ledsign.hardware import LEDSignHardware,LEDSignSelector
from ledsign.keypoint_list import LEDSignKeypoint
from ledsign.program import LEDSignProgramError,LEDSignProgram,LEDSignProgramBuilder
from ledsign.program_io import LEDSignCompiledProgram
from ledsign.protocol import LEDSignUnsupportedProtocolError



__all__=["LEDSign","LEDSignAccessError","LEDSignCompiledProgram","LEDSignDeviceInUseError","LEDSignDeviceNotFoundError","LEDSignHardware","LEDSignKeypoint","LEDSignProgram","LEDSignProgramBuilder","LEDSignProgramError","LEDSignProtocolError","LEDSignSelector","LEDSignUnsupportedProtocolError"]
